#include <amsclil1/amsclil1.h>

int amsclil1_fb32_cmp(uint32_t* a, uint32_t* b, long blen)
{
    int ret = 0;
    long I;

    for(I=blen-1;I>=0;I--)
    {
        if(a[I]>b[I])
        {
            ret = 1;
            break;
        }
        if(b[I]>a[I])
        {
            ret = 2;
            break;
        }
    }

    return ret;
}

int amsclil1_fb32_iszero(uint32_t* a, long blen)
{
    int ret = 1;
    long I;
    for(I=0;I<blen;I++)
    {
        if(a[I]!=0)
        {
            ret = 0;
            break;
        }
    }

    return ret;
}

void amsclil1_fb32_liprint(uint32_t* a, long blen)
{
    long I;
    printf("0x");
    for(I=blen-1;I>=0;I--)
    {
        printf("%08x",a[I]);
    }
    return;
}

void amsclil1_fb16_liprint(uint16_t* a, long blen)
{
    long I;
    printf("0x");
    for(I=blen-1;I>=0;I--)
    {
        printf("%04x",a[I]);
    }
    return;
}

void amsclil1_fb32_libinprint(uint32_t* a, long blen)
{
    long I;
    int J;
    uint32_t bit;
    printf("0b:");
    for(I=blen-1;I>=0;I--)
    {
        for(J=32-1;J>=0;J--)
        {
            bit = (a[I] & (1<<J))>>J;
            printf("%d",bit);
        }
        if(I!=0) printf(":");
    }
}

//bit-shifted reads and writes
//don't move the array in memory, move the read/write
//perhaps it will save time within the division operation
int amsclil1_fb32_readshift(uint32_t* arr, long ind, long shift, uint32_t* val, long blen)
{
    int ret = 0;
    //long Itop = ind + amsc_ltdiv(-shift,32)+1;
    //long Ibot = ind + amsc_ltdiv(-shift,32);
    long Itop = amsc_ltdiv((ind+1)*32-shift,32);
    long Ibot = amsc_ltdiv(ind*32-shift,32);
    int smin = amsc_lmod(-shift,32);

    //printf("\ndebug: %ld, %ld, %ld, %ld, %d\n",ind, shift, Itop, Ibot, smin);

    uint64_t w1,w2;

    *val = 0;
    if(Itop>=0 && Itop<blen)
    {
        w1 = arr[Itop];
        w1 = ((w1 << 32) >> smin); //why does this 32 bit shift work?
        *val = *val | (uint32_t) w1;
    }
    if(Ibot>=0 && Ibot<blen)
    {
        w1 = arr[Ibot];
        w1 = w1 >> smin;
        *val = *val | (uint32_t) w1;
    }

    return ret;
}

int amsclil1_fb32_writeshift(uint32_t* arr, long ind, long shift, uint32_t* val, long blen)
{
    int ret = 0;
    long Itop = amsc_ltdiv((ind+1)*32+shift,32);
    long Ibot = amsc_ltdiv(ind*32+shift,32);
    int smin = amsc_lmod(shift,32);
    int sminc = 32 - smin;

    uint64_t w1,w2;

    //weird problems with shift 32, even shifting uint64_t data types
    //while theres a lot of RTFM garbage online, and everyone will roast you
    //for shifting a type such as uint32_t 32 bits, no one will admit that
    //shifting uint64_t by 32 bits, which by the standard should be defined
    //behavior, fails on x86 processors. Some register is 32 bits wide, and
    //so this doesn't work.

    if(Itop>=0 && Itop<blen)
    {
        if(smin>0) w1 = (*val) >> (32-smin);
        else w1 = 0;
        //w2 = (arr[Itop] >> smin) << smin;
        w2 = (uint64_t) arr[Itop];
        w2 = w2 & (0x00000000ffffffff << smin);
        arr[Itop] = (uint32_t) w1 | (uint32_t) w2;
    }
    if(Ibot>=0 && Ibot<blen)
    {
        w1 = (*val) << smin;
        if(smin<32)
        {
            w2 = (uint64_t) arr[Ibot];
            w2 = w2 & (0xffffffffffffffff - (0x00000000ffffffff << smin));
        }
        else w2 = (uint64_t) arr[Ibot];
        arr[Ibot] = (uint32_t) w1 | (uint32_t) w2;
    }

    return ret;
}

//bit shift operations
//lshift: <<: *2^shift
//wrk[blen]
void amsclil1_fb32_shiftl(uint32_t* a, long shift, uint32_t* wrk, long blen)
{
    long smaj;
    int smin;
    long I;
    uint64_t w1;

    if(shift<0)
    {
        amsclil1_fb32_shiftr(a,-shift,wrk,blen);
    }
    else
    {
        smaj = amsc_ltdiv(shift,32);
        smin = amsc_lmod(shift,32);
        amsc_ui32_memcpy(wrk,a,blen);
        amsc_ui32_memset(a,0,blen);
        for(I=0;I<blen;I++)
        {
            if(I-smaj>=0)
            {
                w1 = (uint64_t) wrk[I-smaj];
                w1 = (w1 << smin) & 0x00000000ffffffff;
                a[I] = a[I] | (uint32_t) w1;
            }
            if(I-smaj-1>=0)
            {
                w1 = (uint64_t) wrk[I-smaj-1];
                w1 = ((w1 << smin) & 0xffffffff00000000);
                w1 = (w1 >> 16) >> 16;
                a[I] = a[I] | (uint32_t) w1;
            }
        }
    }

}


void amsclil1_fb32_shiftr(uint32_t* a, long shift, uint32_t* wrk, long blen)
{
    long smaj;
    int smin;
    long I;
    uint64_t w1;

    if(shift<0)
    {
        amsclil1_fb32_shiftl(a,-shift,wrk,blen);
    }
    else
    {
        smaj = amsc_ltdiv(shift,32);
        smin = amsc_lmod(shift,32);
        amsc_ui32_memcpy(wrk,a,blen);
        amsc_ui32_memset(a,0,blen);
        for(I=0;I<blen;I++)
        {
            if(I+smaj<blen)
            {
                w1 = (uint64_t) wrk[I+smaj];
                w1 = (w1 >> smin);
                a[I] = a[I] | (uint32_t) w1;
            }
            if(I+smaj+1<blen)
            {
                w1 = (uint64_t) wrk[I+smaj+1];
                w1 = ((w1 << 16) << 16) >> smin;
                a[I] = a[I] | (uint32_t) w1;
            }
        }
    }
    return;
}

void amsclil1_ui32_add(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *car)
{
    uint64_t a,b,c;
    a = (uint64_t) op1;
    b = (uint64_t) op2;
    c = a+b;
    *res = (uint32_t) c;
    *car = (uint32_t) ((c>>16)>>16);
    return;
}

void amsclil1_ui32_sub(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *car)
{
    uint64_t a;
    if(op1>=op2)
    {
        *res = op1-op2;
        *car = 0; 
    }
    else
    {
        a = (0x0000000100000000 + (uint64_t) op1) - (uint64_t) op2;
        *res=(uint32_t) a;
        *car = 1;
    }
    return;
}

void amsclil1_ui32_mult(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *car)
{
    uint64_t a,b,c;
    a = (uint64_t) op1;
    b = (uint64_t) op2;
    c = a*b;
    *res = (uint32_t) c;
    *car = (uint32_t) ((c>>16)>>16);
    return;
}

void amsclil1_ui32_div(uint32_t op1, uint32_t op2, uint32_t *res, uint32_t *rem)
{
    if(op2>op1)
    {
        *res = 0;
        *rem = op1;
    }
    else
    {
        *res = op1 / op2;
        *rem = op1 % op2;
    }
    
    return;
}

long amsclil1_fb32_msi(uint32_t*a, long blen)
{
    long ret = 0;
    long I;
    for(I=blen-1;I>=0;I--)
    {
        if(a[I] != 0)
        {
            ret = I;
            break;
        }
    }
    return ret;
}
