#include <amsclil1/amsclil1.h>

//a[blen]
//b[blen]
//c[blen]
void amsclil1_fb32_add(uint32_t *a, uint32_t *b, uint32_t *c, long blen)
{
    long I;
    uint32_t w,c1,c2;

    w = 0;
    c1 = 0;
    c2 = 0;
    amsc_ui32_memset(c,0,blen);
    for(I=0;I<blen;I++)
    {
        amsclil1_ui32_add(a[I],c1,&w,&c1);
        amsclil1_ui32_add(b[I],w,&w,&c2);
        c[I] = w;
        c1 = c1 + c2;
        c2 = 0;
    }

    return;
}

//a[blen]
//b[blen]
//c[blen]
void amsclil1_fb32_sub(uint32_t *a, uint32_t *b, uint32_t *c, long blen)
{
    long I;
    uint32_t w,c1,c2;

    w = 0;
    c1 = 0;
    c2 = 0;
    amsc_ui32_memset(c,0,blen);

    if(amsclil1_fb32_cmp(a,b,blen)==1)
    {
        for(I=0;I<blen;I++)
        {
            amsclil1_ui32_sub(a[I],c1,&w,&c1);
            amsclil1_ui32_sub(w,b[I],&w,&c2);
            c[I] = w;
            c1 = c1 + c2;
            c2 = 0;
        }
    }

    return;
}

//a[blen]
//b[blen]
//c[2*blen]
//wrk - basic mult doesn't need it, but left in in case other schemes are implemented
//      such as Karatsuba's algorithm
//blen - buffer length
void amsclil1_fb32_mult(uint32_t *a, uint32_t *b, uint32_t *c, uint32_t *wrk, long blen)
{
    long I,J;
    long msia, msib;
    uint64_t carry;
    uint64_t f;

    //pull in some of these functions to avoid stack overhead

    for(I=0;I<2*blen;I++) c[I] = 0;
    
    msia = 0;
    for(I=blen-1;I>=0;I--)
    {
        if(a[I]!=0)
        {
            msia = I; break;
        }
    }

    msib = 0;
    for(I=blen-1;I>=0;I--)
    {
        if(b[I]!=0)
        {
            msib = I; break;
        }
    }


    for(I=0;I<=msia;I++)
    {
        f = (uint64_t) a[I];
        carry = 0;
        for(J=0;J<=msib;J++)
        {
            carry = carry + c[I+J] + ((uint64_t)b[J])*f;
            //printf("I:%ld J:%ld a:%08x b:%08x\n",I,J,a[I],b[J]);
            c[I+J] = (uint32_t) (carry & 0x00000000ffffffff);
            carry = (carry >> 16) >> 16;
        }
        if(carry!=0 && ((I+J)<2*blen)) c[I+J] = (uint32_t) (carry & 0x00000000ffffffff);
    }


    return;
}

void amsclil1_fb16_mult(uint16_t *a, uint16_t *b, uint16_t *c, long blen)
{
    long I,J;
    long msia, msib;
    uint32_t carry;
    uint32_t f;

    //pull in some of these functions to avoid stack overhead

    for(I=0;I<2*blen;I++) c[I] = 0;
    
    for(I=blen-1;I>=0;I--)
    {
        if(a[I]!=0)
        {
            msia = I; break;
        }
    }

    for(I=blen-1;I>=0;I--)
    {
        if(b[I]!=0)
        {
            msib = I; break;
        }
    }


    for(I=0;I<=msia;I++)
    {
        f = (uint64_t) a[I];
        carry = 0;
        for(J=0;J<=msib;J++)
        {
            carry = carry + c[I+J] + ((uint32_t)b[J])*f;
            //printf("I:%ld J:%ld a:%08x b:%08x\n",I,J,a[I],b[J]);
            c[I+J] = (uint16_t) (carry & 0x0000ffff);
            carry = (carry >> 16);
        }
        if(carry!=0 && (I+J<2*blen)) c[I+J] = (uint16_t) (carry & 0x0000ffff);
    }


    return;
}

int amsclil1_fb32_shiftcmp(uint32_t *a, uint32_t *b, long bshift, long blen)
{
    int ret = 0;
    long I;
    uint32_t r;
    for(I=blen-1;I>=0;I--)
    {
        amsclil1_fb32_readshift(b,I,bshift,&r,blen);
        if(a[I]>r)
        {
            ret = 1;
            break;
        }
        if(a[I]<r)
        {
            ret = 2;
            break;
        }
    }
    return ret;
}

long amsclil1_fb32_shiftmsi(uint32_t *a, long shift, long blen)
{
    long ret = 0;
    long I;
    uint32_t r;
    for(I=blen-1;I>=0;I--)
    {
        amsclil1_fb32_readshift(a,I,shift,&r,blen);
        if(r!=0)
        {
            ret = I;
            break;
        }
    }

    return ret;
}

void amsclil1_fb32_shiftsub(uint32_t *a, uint32_t *b, long bshift, uint32_t *c, long blen)
{
    long I;
    uint32_t w,c1,c2,r;

    
    amsc_ui32_memset(c,0,blen);

    if(amsclil1_fb32_shiftcmp(a,b,bshift,blen)==1)
    {
        w = 0;
        c1 = 0;
        c2 = 0;
        for(I=0;I<blen;I++)
        {
            //printf("debug:I=%ld a[I]=%ud, c1=%ud\n",I,a[I],c1);
            amsclil1_ui32_sub(a[I],c1,&w,&c1);
            //printf("debug:I=%ld w=%ud, c1=%ud\n",I,w,c1);
            amsclil1_fb32_readshift(b,I,bshift,&r,blen);
            //printf("debug:I=%ld w=%ud, r=%ud\n",I,w,r);
            amsclil1_ui32_sub(w,r,&w,&c2);
            //printf("debug:I=%ld w=%ud, c2=%ud\n",I,w,c2);
            c[I] = w;
            c1 = c1 + c2;
            c2 = 0;
            
        }
    }

    return;
}

void amsclil1_fb32_setzero(uint32_t *a, long blen)
{
    long I;
    for(I=0;I<blen;I++)
    {
        a[I] = 0;
    }
    return;
}

void amsclil1_fb32_shiftliprint(uint32_t* a, long shift, long blen)
{
    long I;
    uint32_t r;

    printf("0x");
    for(I=blen-1;I>=0;I--)
    {
        amsclil1_fb32_readshift(a,I,shift,&r,blen);
        printf("%08x",r);
    }
    return;
}

//internal digit shift comparison
//this is supposed to be a <=> b*base^bshft
int amsclil1_fb16_intlscmp(uint16_t *a, uint16_t *b, long bshft, long blen)
{
    int cmp = 0;
    long I;
    uint16_t r1,r2;

    //cover the whole range, dont skip higher digits of b
    for(I=(blen-1)+bshft;I>=0;I--)
    {
        //bug found: >=0, not >0
        if((I-bshft>=0) && (I-bshft<blen)) r1 = b[I-bshft]; else r1=0;
        if((I<blen) && (I>=0)) r2=a[I]; else r2=0; //probably >=0

        if(r2>r1) {cmp=1; break;}
        if(r2<r1) {cmp=2; break;}
    }


    //think about this - I think there is a bug
    // for(I=blen-1;I>=0;I--)
    // {
    //     if(I-bshft>=0) r = b[I-bshft]; else r=0;
    //     if(a[I]>r) {cmp=1; break;}
    //     if(a[I]<r) {cmp=2; break;}
    // }
    return cmp;
}

//internal digit shift subtraction
//this is supposed to be something like c = a - b*base^bshft
void amsclil1_fb16_intlssub(uint16_t *a, uint16_t *b, long bshft, uint16_t *c, long blen)
{
    long I;
    int cmp = amsclil1_fb16_intlscmp(a,b,bshft,blen);
    uint32_t w0,w1,car;
    
    for(I=0;I<blen;I++) c[I]=0;

    if(cmp!=1)
    {
        //for(I=0;I<blen;I++) c[I]=0;
    }
    else
    {
        car = 0;
        for(I=0;I<bshft;I++) c[I] = a[I];
        for(I=bshft;I<blen;I++)
        {
            w0 = (uint32_t) a[I];
            w1 = (uint32_t) b[I-bshft]+car;
            if(w0>=w1) {c[I] = (uint16_t) (w0-w1); car=0;}
            else {c[I] = (uint16_t) (0x00010000+w0-w1); car=1;}
        }
    }
    
    return;
}

//note - blen must be one greater than msia to contain answer under all circumstances
void amsclil1_fb16_intlmultsing(uint16_t *a, uint16_t b, long blen)
{
    long I;
    uint32_t car;
    car = 0;
    for(I=0;I<blen;I++)
    {
        car = car + ((uint32_t)a[I])*((uint32_t) b);
        a[I] = (uint16_t) (car&0x0000ffff);
        car = car >> 16;
    }

    return;
}

//Division using the algorithm provided in 
//Handbook of Applied Cryptography, Chapter 14 (Algorithm 14.20)
//I messed this up, and need to try to redo it
// there is something screwy going on where 
//x [blen]
//y [blen]
//res [blen]
//rem [blen]
//wrk [5*(blen+1)]
//blen - length of integer buffers
//
//  Updated 8 Nov 22 - fixed some bugs
// - why can step 2 not be *entirely* skipped if we just adapt step 3, starting with x[msix+1]? 
int amsclil1_fb32_div_v3(uint32_t *x, uint32_t *y, uint32_t *res, uint32_t *rem, uint32_t* wrk, long blen)
{
    int ret = 0;
    int rr = 0;

    long I,J;

    uint16_t* x2 = (uint16_t*) &wrk[0];
    uint16_t* y2 = (uint16_t*) &wrk[1*(blen+1)];
    uint16_t* q2 = (uint16_t*) &wrk[2*(blen+1)];
    uint16_t* r2 = (uint16_t*) &wrk[3*(blen+1)];
    uint16_t* r3 = (uint16_t*) &wrk[4*(blen+1)];
    uint16_t* tmp;
    int dbg;
    

    long N2 = 2*blen; //length of uint16 integers

    long msix2,msiy2; //most significant index of x and y

    //intiialize vars
    for(I=0;I<N2;I++)
    {
        x2[I] = ((uint16_t*)x)[I];
        y2[I] = ((uint16_t*)y)[I];
        q2[I] = 0;
        r2[I] = 0;
        r3[I] = 0;
    }
    x2[N2] = 0;
    y2[N2] = 0;
    q2[N2] = 0;
    r2[N2] = 0;
    r3[N2] = 0;
    x2[N2+1] = 0;
    y2[N2+1] = 0;
    q2[N2+1] = 0;
    r2[N2+1] = 0;
    r3[N2+1] = 0;
    //(0 to blen-1) initialized

    //intiialize outputs
    for(I=0;I<blen;I++)
    {
        res[I] = 0;
        rem[I] = 0;
    }

    //find most significant index of x2 and y2
    msix2 = 0;
    for(I=N2-1;I>=0;I--)
    {
        if(x2[I]!=0)
        {
            msix2 = I;
            break;
        }
    }

    msiy2 = 0;
    for(I=N2-1;I>=0;I--)
    {
        if(y2[I]!=0)
        {
            msiy2 = I;
            break;
        }
    }

    //if y==0
    if(msiy2==0 && y2[0]==0)
    {
        ret = 0;
        return ret;
    }

    //if x == 0
    if(msix2==0&&x2[0]==0)
    {
        ret = 1;
        return ret;
    }

    //if y>x ... 1
    if(msiy2>msix2)
    {
        ret = 1;
        for(I=0;I<blen;I++)
        {
            rem[I] = x[I];
        }
        return ret;
    }

    //if y>x ... 2
    if(msiy2==msix2)
    {
        for(I=msix2;I>=0;I--)
        {
            if(x2[I]>y2[I])
            {
                break;
            }
            if(y2[I]>x2[I])
            {
                ret = 1;
                for(I=0;I<blen;I++)
                {
                    rem[I] = x[I];
                }
                return ret;
            }
        }
    }

    //algorithm is only valid for n>=t>=1
    //if msix2==misy2==0, then do single digit division
    if(msix2==0)
    {
        ret =  1;
        res[0] = (uint32_t) (x2[0]/y2[0]);
        rem[0] = (uint32_t) (x2[0] - y2[0]*((uint16_t)res[0]));
        return ret;
    }
    
    /// Start Over ///
    //Refer to HAC 14.20

    // In version 3, I'm going to try skipping step two, and starting at msix2+1 at step 3.
    // I have room for it in memory, and x2[msix2+1] should be 0.

    uint64_t bv0,bv1,bv2;

    //3: for I from N down to (T+1) do...
    for(I=msix2+1;I>=msiy2+1;I--)
    {

        //3.1: if xi=yt q_i-t-1 = base-1; else qi-t-1 = (xi*base+xi-1)/y_t

        //Modify 3.1: q_i-t-1 = (x_i*b^2+x_i-1*b+x_i-2)/(yt*b+y_t-1)
        if(x2[I]==y2[msiy2])
        {
            q2[I-msiy2-1] = (uint16_t)(0xffff);
        }
        else
        {
            //I is guaranteed to be >=1
            bv0 = (((uint64_t)x2[I])<<16LL) + (uint64_t)x2[I-1];
            bv1 = ((uint64_t) y2[msiy2]);

            bv2 = bv0/bv1;
            q2[I-msiy2-1] = (int16_t) (bv2 & 0x000000000000ffffLL);
        }

        //3.2: while (expression > expression) do ...
        dbg = 0;

        bv0 = ((uint64_t)q2[I-msiy2-1])*( (((uint64_t)y2[msiy2])<<16LL) + (uint64_t)y2[msiy2-1] ); 
        //printf("debug: 0x%lx 0x%lx 0x%lx\n",((uint64_t)q2[I-msiy2-1]),(((uint64_t)y2[msiy2])<<16LL),(uint64_t)y2[msiy2-1]);

        bv1 = (((uint64_t)x2[I])<<32LL)+(((uint64_t)x2[I-1])<<16LL)+((uint64_t)x2[I-2]);
        //printf("bv0=0x%lx, bv1=0x%lx\n",bv0,bv1);
        while(bv0>bv1)
        {
            //printf("%d %llx %llx %llx ",dbg,(long long unsigned)q2[I-msiy2-1],(long long unsigned)bv0,(long long unsigned)bv1);
            //printf("%x %x %x \n",(int)x2[I],(int)x2[I-1],(int)x2[I-2]);

            q2[I-msiy2-1] = q2[I-msiy2-1]-1;
            bv0 = ((uint64_t)q2[I-msiy2-1])*( (((uint64_t)y2[msiy2])<<16LL) + (uint64_t)y2[msiy2-1] );
            bv1 = (((uint64_t)x2[I])<<32LL)+(((uint64_t)x2[I-1])<<16LL)+((uint64_t)x2[I-2]);

            //printf("bv0=0x%lx, bv1=0x%lx\n",bv0,bv1);
            //printf("debug: next q[%ld] = %d\n",I-msiy2-1,q2[I-msiy2-1]);
            
            dbg++;
        }

        //3.3 and 3.4

        //r3 = q[I-t-1]*y2*base^(I-t-1)
        for(J=0;J<N2+2;J++)
        {
            r3[J] = y2[J];
        }
        amsclil1_fb16_intlmultsing(r3,q2[I-msiy2-1],N2+1);

        
        if(amsclil1_fb16_intlscmp(x2,r3,I-msiy2-1,N2+1)==2)
        {
            //if r3>x

            //q[I-t-1] = q[I-t-1]-1
            dbg = 0;
            while(amsclil1_fb16_intlscmp(x2,r3,I-msiy2-1,N2+1)==2)
            {
                q2[I-msiy2-1] = q2[I-msiy2-1]-1;
                //recalc r3
                for(J=0;J<N2+2;J++)
                {
                    r3[J] = y2[J];
                }
                amsclil1_fb16_intlmultsing(r3,q2[I-msiy2-1],N2+1);
                //could also subtract y2 here (speed?)
                dbg++;
            }
            //printf("debug: step3.4 %d iterations\n",dbg);

            //this step seems like it must be repeated until x2<r3
            //but in HAC, it seems that it is never applied more than once.
            
            //x = x - r3
            amsclil1_fb16_intlssub(x2,r3,I-msiy2-1,r2,N2+1);
            tmp = x2; x2 = r2; r2 = tmp;
        }
        else
        {
            //if r3<=x

            //x = x - q[I-t-1]*y*base^(I-t-1)

            amsclil1_fb16_intlssub(x2,r3,I-msiy2-1,r2,N2+1);
            tmp = x2; x2 = r2; r2 = tmp;
        }
    } // for I = msixy; I>msiy2+1; I--

    //rem = x
    //res = q
    ret = 1;
    for(I=0;I<N2;I++)
    {
        ((uint16_t*)rem)[I] = x2[I];
        ((uint16_t*)res)[I] = q2[I];
    }

    return ret;
}

//Division using the algorithm provided in 
//Handbook of Applied Cryptography, Chapter 14 (Algorithm 14.20)
//I messed this up, and need to try to redo it
// there is something screwy going on where 
//x [blen]
//y [blen]
//res [blen]
//rem [blen]
//wrk [5*(blen+1)]
//blen - length of integer buffers
//
//  Updated 8 Nov 22 - fixed some bugs
// - why can step 2 not be *entirely* skipped if we just adapt step 3, starting with x[msix+1]? 
int amsclil1_fb32_div_v2(uint32_t *x, uint32_t *y, uint32_t *res, uint32_t *rem, uint32_t* wrk, long blen)
{
    int ret = 0;
    int rr = 0;

    long I,J;

    uint16_t* x2 = (uint16_t*) &wrk[0];
    uint16_t* y2 = (uint16_t*) &wrk[1*(blen+1)];
    uint16_t* q2 = (uint16_t*) &wrk[2*(blen+1)];
    uint16_t* r2 = (uint16_t*) &wrk[3*(blen+1)];
    uint16_t* r3 = (uint16_t*) &wrk[4*(blen+1)];
    uint16_t* tmp;
    int dbg;
    

    long N2 = 2*blen; //length of uint16 integers

    long msix2,msiy2; //most significant index of x and y

    //intiialize vars
    for(I=0;I<N2;I++)
    {
        x2[I] = ((uint16_t*)x)[I];
        y2[I] = ((uint16_t*)y)[I];
        q2[I] = 0;
        r2[I] = 0;
        r3[I] = 0;
    }
    x2[N2] = 0;
    y2[N2] = 0;
    q2[N2] = 0;
    r2[N2] = 0;
    r3[N2] = 0;
    x2[N2+1] = 0;
    y2[N2+1] = 0;
    q2[N2+1] = 0;
    r2[N2+1] = 0;
    r3[N2+1] = 0;
    //(0 to blen-1) initialized

    //intiialize outputs
    for(I=0;I<blen;I++)
    {
        res[I] = 0;
        rem[I] = 0;
    }

    //find most significant index of x2 and y2
    msix2 = 0;
    for(I=N2-1;I>=0;I--)
    {
        if(x2[I]!=0)
        {
            msix2 = I;
            break;
        }
    }

    msiy2 = 0;
    for(I=N2-1;I>=0;I--)
    {
        if(y2[I]!=0)
        {
            msiy2 = I;
            break;
        }
    }

    //if y==0
    if(msiy2==0 && y2[0]==0)
    {
        ret = 0;
        return ret;
    }

    //if x == 0
    if(msix2==0&&x2[0]==0)
    {
        ret = 1;
        return ret;
    }

    //if y>x ... 1
    if(msiy2>msix2)
    {
        ret = 1;
        for(I=0;I<blen;I++)
        {
            rem[I] = x[I];
        }
        return ret;
    }

    //if y>x ... 2
    if(msiy2==msix2)
    {
        for(I=msix2;I>=0;I--)
        {
            if(x2[I]>y2[I])
            {
                break;
            }
            if(y2[I]>x2[I])
            {
                //printf("debug: check1\n");
                ret = 1;
                for(I=0;I<blen;I++)
                {
                    rem[I] = x[I];
                }
                return ret;
            }
        }
    }

    //algorithm is only valid for n>=t>=1
    //if msix2==misy2==0, then do single digit division
    if(msix2==0)
    {
        ret =  1;
        res[0] = (uint32_t) (x2[0]/y2[0]);
        rem[0] = (uint32_t) (x2[0] - y2[0]*((uint16_t)res[0]));
        return ret;
    }
    
    /// Start Over ///
    //Refer to HAC 14.20

    //2: find initial value for first digit
    //while (x>y*b^(n-t))
    // this block is suspect - I don't understand what it's doing - why doesn't it start from a guessed value?
    // it iterates, a *lot*.

    //initial q guess = x_t/y_n - 1
    
    //printf("dbg: x2="); amsclil1_fb16_liprint(x2,N2+2); printf("\n");
    //printf("dbg: y2="); amsclil1_fb16_liprint(y2,N2+2); printf("\n");
    //printf("dbg: msix2=%ld msiy2=%ld\n",msix2,msiy2);

    q2[msix2-msiy2] = x2[msix2]/y2[msiy2];
    for(I=0;I<=N2+1;I++) //r3 = y2
    {
        if(I<=msiy2) r3[I] = y2[I];
        else r3[I] = 0;
    }
    //printf("dbg: r3="); amsclil1_fb16_liprint(r3,N2+2); printf("\n");
    // //r3 = y*q[n-t]
    amsclil1_fb16_intlmultsing(r3,q2[msix2-msiy2],N2+1);
    //printf("dbg: r3="); amsclil1_fb16_liprint(r3,N2+2); printf("\n");
    
    //updated initial guess iteration
    //while (r3>x)
    //{ r3 = r3 - y; q = q - 1 }

    

    while(amsclil1_fb16_intlscmp(x2,r3,msix2-msiy2,N2+1)==2)
    {
        amsclil1_fb16_intlssub(r3,y2,0,r2,N2+1);
        tmp = r2; r2 = r3; r3 = tmp;
        q2[msix2-msiy2] = q2[msix2-msiy2] - 1;

        //printf("q2=%d\n",q2[msix2-msiy2]);
        //printf("x2="); amsclil1_fb16_liprint(x2,N2+2); printf("\n");
        //printf("r3="); amsclil1_fb16_liprint(r3,N2+2); printf("\n");
    }
    //x2 = x2 - r3
    amsclil1_fb16_intlssub(x2,r3,msix2-msiy2,r2,N2+1);
    tmp = x2; x2 = r2; r2 = tmp;

    //printf("x2="); amsclil1_fb16_liprint(x2,N2+2); printf("\n");

    //printf("debug: q2[%ld]=%d\n",msix2-msiy2,q2[msix2-msiy2]);

    //old initial guess
    // dbg = 0;
    // while(amsclil1_fb16_intlscmp(x2,y2,msix2-msiy2,N2)!=2)
    // {
    //     q2[msix2-msiy2] = q2[msix2-msiy2]+1;
        
    //     // x = x - y*b^(n-t)
    //     amsclil1_fb16_intlssub(x2,y2,msix2-msiy2,r3,N2); // this might also be suspect
    //     tmp = x2; x2 = r3; r3 = tmp; //swap(x2,r3)
        

    //     //printf("dbg: x2="); amsclil1_fb16_liprint(x2,N2+2); printf("\n");
    //         // this is not firing for some reason
    //         // amsclil1_fb16_intlscmp is suspect
            
    //     // if(dbg>=100)
    //     // {
    //     //     printf("debug1: dbg>100\n");
    //     //     exit(0);
    //     // }
    //     // dbg++;
    // }

    // End step 2

    uint64_t bv0,bv1,bv2;

    //3: for I from N down to (T+1) do...
    for(I=msix2;I>=msiy2+1;I--)
    {

        //3.1: if xi=yt q_i-t-1 = base-1; else qi-t-1 = (xi*base+xi-1)/y_t

        //Modify 3.1: q_i-t-1 = (x_i*b^2+x_i-1*b+x_i-2)/(yt*b+y_t-1)
        if(x2[I]==y2[msiy2])
        {
            q2[I-msiy2-1] = (uint16_t)(0xffff);
        }
        else
        {
            //I is guaranteed to be >=1
            bv0 = (((uint64_t)x2[I])<<16LL) + (uint64_t)x2[I-1];
            bv1 = ((uint64_t) y2[msiy2]);

            //bv1 = ((uint64_t)q2[I-msiy2-1])*( (((uint64_t)y2[msiy2])<<16LL) + (uint64_t)y[msiy2-1] );
            //bv0 = (((uint64_t)x2[I])<<32LL)+(((uint64_t)x2[I-1])<<16LL)+((uint64_t)x2[I-2]);
            bv2 = bv0/bv1;
            q2[I-msiy2-1] = (int16_t) (bv2 & 0x000000000000ffffLL);
        }

        //printf("debug: initial q[%ld] = %d\n",I-msiy2-1,q2[I-msiy2-1]);

        //3.2: while (expression > expression) do ...
        dbg = 0;

        bv0 = ((uint64_t)q2[I-msiy2-1])*( (((uint64_t)y2[msiy2])<<16LL) + (uint64_t)y2[msiy2-1] ); 
            //this expression is wrong for some reason - y! not y2! needs to be y2!!!!
        //printf("debug: 0x%lx 0x%lx 0x%lx\n",((uint64_t)q2[I-msiy2-1]),(((uint64_t)y2[msiy2])<<16LL),(uint64_t)y2[msiy2-1]);

        bv1 = (((uint64_t)x2[I])<<32LL)+(((uint64_t)x2[I-1])<<16LL)+((uint64_t)x2[I-2]);
        //printf("bv0=0x%lx, bv1=0x%lx\n",bv0,bv1);
        while(bv0>bv1)
        {
            //printf("%d %llx %llx %llx ",dbg,(long long unsigned)q2[I-msiy2-1],(long long unsigned)bv0,(long long unsigned)bv1);
            //printf("%x %x %x \n",(int)x2[I],(int)x2[I-1],(int)x2[I-2]);

            q2[I-msiy2-1] = q2[I-msiy2-1]-1;
            bv0 = ((uint64_t)q2[I-msiy2-1])*( (((uint64_t)y2[msiy2])<<16LL) + (uint64_t)y2[msiy2-1] );
            bv1 = (((uint64_t)x2[I])<<32LL)+(((uint64_t)x2[I-1])<<16LL)+((uint64_t)x2[I-2]);

            //printf("bv0=0x%lx, bv1=0x%lx\n",bv0,bv1);
            //printf("debug: next q[%ld] = %d\n",I-msiy2-1,q2[I-msiy2-1]);
            
            dbg++;
        }
        //printf("debug: step3.2: %d iterations\n",dbg);

        //3.3 and 3.4

        //r3 = q[I-t-1]*y2*base^(I-t-1)
        for(J=0;J<N2+2;J++)
        {
            r3[J] = y2[J];
        }
        amsclil1_fb16_intlmultsing(r3,q2[I-msiy2-1],N2+1);

        
        if(amsclil1_fb16_intlscmp(x2,r3,I-msiy2-1,N2+1)==2)
        {
            //if r3>x

            //q[I-t-1] = q[I-t-1]-1
            dbg = 0;
            while(amsclil1_fb16_intlscmp(x2,r3,I-msiy2-1,N2+1)==2)
            {
                q2[I-msiy2-1] = q2[I-msiy2-1]-1;
                //recalc r3
                for(J=0;J<N2+2;J++)
                {
                    r3[J] = y2[J];
                }
                amsclil1_fb16_intlmultsing(r3,q2[I-msiy2-1],N2+1);
                //could also subtract y2 here (speed?)
                dbg++;
            }
            //printf("debug: step3.4 %d iterations\n",dbg);

            //this step seems like it must be repeated until x2<r3
            //but in HAC, it seems that it is never applied more than once.

            //sometimes this is missing r3 being greater
            //x2 already seems to be zero when this condition obtains.
            // printf("x2="); amsclil1_fb16_liprint(x2,N2+2); printf("\n");
            // printf("r3="); amsclil1_fb16_liprint(r3,N2+2); printf("\n");
            // printf("diff=%d\n",(int)(I-msiy2-1));
            // printf("cmp=%d\n",amsclil1_fb16_intlscmp(x2,r3,I-msiy2-1,N2+1));
            // printf("q2[I-t-1]=%x\n",(int)q2[I-msiy2-1]);
            
            //x = x - r3
            amsclil1_fb16_intlssub(x2,r3,I-msiy2-1,r2,N2+1);
            tmp = x2; x2 = r2; r2 = tmp;
        }
        else
        {
            //if r3<=x

            //sometimes this is missing r3 being greater
            //x2 already seems to be zero when this condition obtains.
            // printf("x2="); amsclil1_fb16_liprint(x2,N2+2); printf("\n");
            // printf("r3="); amsclil1_fb16_liprint(r3,N2+2); printf("\n");

            //x = x - q[I-t-1]*y*base^(I-t-1)

            amsclil1_fb16_intlssub(x2,r3,I-msiy2-1,r2,N2+1);
            tmp = x2; x2 = r2; r2 = tmp;
        }
    } // for I = msixy; I>msiy2+1; I--

    //rem = x
    //res = q
    ret = 1;
    for(I=0;I<N2;I++)
    {
        ((uint16_t*)rem)[I] = x2[I];
        ((uint16_t*)res)[I] = q2[I];
    }

    return ret;
}

//Division using the algorithm provided in 
//Handbook of Applied Cryptography, Chapter 14 (Algorithm 14.20)
//a [blen]
//b [blen]
//res [blen]
//rem [blen]
//wrk [5*(blen+1)]
void amsclil1_fb32_div_v0(uint32_t *a, uint32_t *b, uint32_t *res, uint32_t *rem, uint32_t* wrk, long blen)
{
    //There are operations in the algorithm requiring 3-digit types
    //uint32_t must be converted to uint16_t for this algorithm
    long I,J;
    long msia,msib,diff;
    int cmp,izb;

    long N2 = blen*2; //uint16_t length is 2*uint32_t length
    uint16_t* bswap = NULL;
    uint16_t* x2 = (uint16_t*) (&wrk[0]);
    uint16_t* y2 = (uint16_t*) (&wrk[blen+1]);
    uint16_t* q2 = (uint16_t*) (&wrk[2*(blen+1)]);
    uint16_t* r2 = (uint16_t*) (&wrk[3*(blen+1)]);
    uint16_t* r3 = (uint16_t*) (&wrk[4*(blen+1)]);

    uint64_t bv0,bv1,bv2;
    uint64_t bx0,bx1,bx2;
    uint64_t by0,by1,by2;

    long dbg;

    for(I=0;I<N2;I++)
    {
        x2[I] = ((uint16_t*)a)[I];
        y2[I] = ((uint16_t*)b)[I];
        q2[I] = 0;
        r2[I] = 0;
        //printf("I: %ld x: %04x y:%04x \n",I,x2[I],y2[I]);
    }
    for(I=N2;I<N2+2;I++)
    {
        x2[I] = 0;
        y2[I] = 0;
        q2[I] = 0;
        r2[I] = 0;
    }

    //printf("check1\n");

    msia = 0;
    for(I=N2-1;I>=0;I--)
    {
        if(x2[I]!=0)
        {
            msia = I; break;
        }
    }

    msib = 0;
    for(I=N2-1;I>=0;I--)
    {
        if(y2[I]!=0)
        {
            msib = I; break;
        }
    }

    //printf("check2\n");

    cmp = 0;
    for(I=N2-1;I>=0;I--)
    {
        if(x2[I]>y2[I]) {cmp = 1; break;}
        if(y2[I]>x2[I]) {cmp = 2; break;}
    }

    //printf("check3\n");

    izb = 1;
    for(I=0;I<N2;I++)
    {
        if(y2[I]!=0) {izb = 0; break;}
    }

    //printf("check4\n");
    //printf("debug: %ld %ld %d\n",msia,msib,izb);

    if(izb)
    {
        for(I=0;I<blen;I++) {res[I] = 0; rem[I] = 0;}
    }
    else if(cmp==0)
    {
        for(I=0;I<blen;I++) {res[I] = 0; rem[I] = 0; res[0]=1;}
    }
    else if(cmp==2)
    {
        for(I=0;I<blen;I++) {rem[I] = a[I]; res[I] = 0;}
    }
    else if(msib < 1)
    {
            //do I need to handle these differently? yes - the below process hangs when msib<1
            //printf("too few sigfigs - unimplemented.\n");

            //single digit division
            bv0 = 0;
            for(I=msia;I>=0;I--)
            {
                bv0 = bv0 + ((uint64_t)x2[I]);
                bv1 = bv0/((uint64_t)y2[0]);
                bv0 = (bv0 % ((uint64_t)y2[0])) << 16;
                q2[I] = (uint16_t) (bv1 & 0x000000000000ffff);
            }
            r2[0] = (uint16_t) (bv0>>16);

            for(I=0;I<N2;I++)
            {
                ((uint16_t*)res)[I] = q2[I];
                ((uint16_t*)rem)[I] = r2[I];
            }
    }
    else
    {
        //perform the division
        diff = msia-msib+1;
        //printf("debug: diff=%ld\n",diff);
        //first digit subtraction
        I = msia;
        //shifted cmp: x>=y*b^(diff-1)
        //printf("check5\n");

        //Note - certain numbers hang before step 6
        // 0x000...a62c4b
        // 0x000...377d3a is an example

        dbg = 0;
        while(amsclil1_fb16_intlscmp(x2,y2,diff-1,N2+2)!=2)
        {
            q2[diff-1] = q2[diff-1]+1;

            // if(dbg>=495 || dbg<=5)
            // {
            //     printf("debug: %ld y2=", diff); amsclil1_fb16_liprint(y2,N2+2); printf("\n");
            //     printf("debug: %ld x2=", diff); amsclil1_fb16_liprint(x2,N2+2); printf("\n");
            // }

            amsclil1_fb16_intlssub(x2,y2,diff-1,r2,N2+2); //x = x - y*b^(diff-1)
            bswap = x2; x2 = r2; r2 = bswap; //swap(x2,r2)
             
             dbg++;
            //  if(dbg>=500)
            //  {
            //      printf("debug: dbg exceeded 500 iterations in loop 1.\n");
            //      exit(0);
            //  }
        }

        //printf("check6\n");

        //next N digits quotient
        by1 = (uint64_t)y2[msib];
        if(msib-1>=0) by0 = (uint64_t)y2[msib-1]; else by0 = 0;

        for(I=msia; I>=msib+1; I--)
        {
            bx2 = (uint64_t)x2[I];
            if(I-1>=0) bx1 = (uint64_t)x2[I-1]; else bx1 = 0;
            if(I-2>=0) bx0 = (uint64_t)x2[I-2]; else bx2 = 0;
            
            bv0 = ((bx2<<16)+bx1)/by1; //3.1
            bv1 = bv0*((by1<<16)+by0); //3.2
            bv2 = ((bx2<<16)<<16)+(bx1<<16)+bx0;

            dbg = 0;
            while(bv1>bv2)
            {
                bv0 = bv0 - 1;
                bv1 = bv0*((by1<<16)+by0);
                bv2 = ((bx2<<16)<<16)+(bx1<<16)+bx0;

                dbg++;
                // if(dbg>=100)
                // {
                //     printf("debug: dbg exceeded 100 iterations in loop 2.\n");
                //     exit(0);
                // }
            }

            //3.3 and 3.4 rearranged a bit
            q2[I-msib-1] = (uint16_t) bv0;
            //n. r2 = q[I-msib-1]*y
            //r2 = r2 << (i-msib-1)
            for(J=0;J<N2;J++) r2[J] = y2[J];
            amsclil1_fb16_intlmultsing(r2,q2[I-msib-1],N2+2);
            cmp = amsclil1_fb16_intlscmp(x2,r2,I-msib-1,N2+2);
            //if r2>x q2 = q2 - 1; repeat step n.
            if(cmp==2)
            {
                q2[I-msib-1]=q2[I-msib-1]-1;
                for(J=0;J<N2;J++) r2[J] = y2[J];
                amsclil1_fb16_intlmultsing(r2,q2[I-msib-1],N2+2);
            }
            //x = x - r2
            amsclil1_fb16_intlssub(x2,r2,I-msib-1,r3,N2+2);
            bswap = x2; x2 = r3; r3 = bswap;
        }
        for(I=0;I<N2;I++)
        {
            ((uint16_t*)res)[I] = q2[I];
            ((uint16_t*)rem)[I] = x2[I];
        }
    }
    
    return;
}


//Division using the algorithm provided in 
//Handbook of Applied Cryptography, Chapter 14 (Algorithm 14.20)
//a [blen]
//b [blen]
//res [blen]
//rem [blen]
//wrk [5*(blen+1)]
int amsclil1_fb32_div(uint32_t *a, uint32_t *b, uint32_t *res, uint32_t *rem, uint32_t* wrk, long blen)
{
    int ret = 0;
    //amsclil1_fb32_div_v0(a,b,res,rem,wrk,blen);
    ret = amsclil1_fb32_div_v3(a,b,res,rem,wrk,blen);
    return ret;
}




//a[blen]
//dd[3*blen]
//wrk[]
//blen - base buffer length

//While this will be much faster with fast divisions, it's still O(N^3) with O(N^2) divisions
//Do it differely - look up Knuth's algorithm
void amsclil1_fb32_decimal(uint32_t *a, uint8_t *dd, uint32_t* wrk, int blen)
{


    return;
}


//calculates mod(a^pw, md)
//a[blen] - base buffer
//pw[blen] - power
//md[blen] - modulus
//wrk[8*(blen+2)]
//stor[12*(blen+1)]
//blen - base buffer length
void amsclil1_fb32_modpower(uint32_t *a, uint32_t *pw, uint32_t *md, uint32_t *res, uint32_t *wrk, uint32_t *stor, int blen)
{

    //Classical Fast Modular Exponentiation Algorithm
    //amsclil1_fb32_modpower_cls(a,pw,md,res,wrk,stor,blen);

    //Fast Modular Exponentiation using Barrett Reduction
    amsclil1_fb32_modpower_bar(a,pw,md,res,wrk,stor,blen);

    return;
}

//Classical Fast Modular Exponentiation Algorithm
//calculates mod(a^pw, md)
//a[blen] - base buffer
//pw[blen] - power
//md[blen] - modulus
//wrk[66*(blen+1)]
//stor[28*(blen+1)]
//blen - base buffer length
void amsclil1_fb32_modpower_cls(uint32_t *a, uint32_t *pw, uint32_t *md, uint32_t *res, uint32_t *wrk, uint32_t *stor, int blen)
{
    long I;
    long msipw;
    uint32_t bit;
    long smaj;
    int smin;

    uint32_t* bswap = NULL;
    uint32_t* bm0 = &(stor[0]);
    uint32_t* bm1 = &(stor[2*(blen+1)]);
    uint32_t* ra = &(stor[4*(blen+1)]);
    uint32_t* rb = &(stor[6*(blen+1)]);
    uint32_t* acc = &(stor[8*(blen+1)]);
    uint32_t* mdb = &(stor[10*(blen+1)]);

    msipw = amsclil1_fb32_msi(pw,blen);

    //init acc buffer to 1
    amsc_ui32_memset(acc,0,2*(blen+1));
    acc[0] = 0x00000001;
    
    //init bm0 to a
    amsc_ui32_memset(bm0,0,2*(blen+1));
    amsc_ui32_memcpy(bm0,a,blen);

    //move mod to size 2*blen+1 buffer
    amsc_ui32_memset(mdb,0,2*(blen+1));
    amsc_ui32_memcpy(mdb,md,blen);

    for(I=0;I<32*(msipw+1);I++)
    {
        smaj = I/32;
        smin = I%32;
        bit = (pw[smaj] & (1<<smin))>>smin;

        if(bit!=0)
        {
            //printf("debug: I=%ld\n",I);
            amsclil1_fb32_mult(acc,bm0,ra,wrk,blen);
            bswap = acc; acc = ra; ra = bswap; //swap(acc,ra)
            //printf("debug: acc="); amsclil1_fb32_liprint(acc,2*blen); printf("\n");
            //amsclil1_fb32_div(acc,mdb,ra,rb,wrk,2*blen+1);
            amsclil1_fb32_div(acc,mdb,ra,rb,wrk,2*blen);
            bswap = acc; acc = rb; rb = bswap; //swap(acc,rb)
            //printf("debug: acc="); amsclil1_fb32_liprint(acc,2*blen); printf("\n");
        }

        //bm0 = (bm0**2) % md
        amsclil1_fb32_mult(bm0,bm0,bm1,wrk,blen);
        amsclil1_fb32_div(bm1,mdb,ra,bm0,wrk,2*blen);
    }

    amsc_ui32_memcpy(res,acc,blen);

    return;
}
