#include <amsclil1/amsclil1.h>


//Euler greatest common divisor algorithm
//a [blen]
//b [blen]
//gcd [blen]
//use division algorithm for modulus - no need for modpower here
//wrk [9*(blen+1)] ....
void amsclil1_fb32_eulergcd(uint32_t *a, uint32_t *b, uint32_t *gcd, uint32_t *wrk, long blen)
{
    long I;
    uint32_t *ai,*bi,*ci,*tmp,*ri,*wrki;

    int dbg;
    
    ai = &(wrk[0*(blen+1)]); //[blen+1]
    bi = &(wrk[1*(blen+1)]); //[blen+1]
    ci = &(wrk[2*(blen+1)]); //[blen+1]
    ri = &(wrk[3*(blen+1)]); //gcd; //[blen]
    wrki = &(wrk[4*(blen+1)]); //5*(blen+1)
    //total 9*(blen+1)

    for(I=0;I<blen+1;I++)
    {
        ci[I] = 0;
    }
    
    // if a>b
    if(amsclil1_fb32_cmp(a,b,blen)!=2)
    {
        //copy a to ai, b to bi
        for(I=0;I<blen;I++)
        {
            ai[I] = a[I];
            bi[I] = b[I];
            ri[I] = 0;
        }
    }
    else
    {
        //copy a to bi, b to ai
        for(I=0;I<blen;I++)
        {
            bi[I] = a[I];
            ai[I] = b[I];
            ri[I] = 0;
        }
    }
    ai[blen] = 0;
    bi[blen] = 0;
    ri[blen] = 0;

    dbg = 0;

    while(amsclil1_fb32_iszero(bi,blen)==0)
    {
        //ri = ai mod bi
        
        //printf("%d pre div.\n", dbg);
        //printf("a="); amsclil1_fb32_liprint(ai,blen); printf("\n");
        //printf("b="); amsclil1_fb32_liprint(bi,blen); printf("\n");

        amsclil1_fb32_div(ai,bi,ci,ri,wrki,blen);
        //printf("post div.\n");

        
        //printf("c="); amsclil1_fb32_liprint(ci,blen); printf("\n");
        //printf("r="); amsclil1_fb32_liprint(ri,blen); printf("\n");
        //printf("\n");

        //a = b, b=r
        tmp = ai;
        ai = bi;
        bi = ri;
        ri = tmp;

        dbg++;
    }
    //printf("post loop");

    //gcd = a
    for(I=0;I<blen;I++)
    {
        gcd[I] = ai[I];
    }

    return;
}



//Least Common Multiple
//
//  returns lcm = lcm(a,b)
//  lcm(0,0) = 0 (special case)
//
//a [blen]
//b [blen]
//lcm [2*blen]
//wrk [14*(blen+1)]
//blen - max number of digits in a and b
//
void amsclil1_fb32_lcm(uint32_t *a, uint32_t *b, uint32_t *lcm, uint32_t *wrk, long blen)
{
    uint32_t *aint; //internal smallest input
    uint32_t *bint; //internal largest input
    uint32_t *gcd = &(wrk[0]);      //2*blen
    uint32_t *q = &(wrk[2*blen]);   //1*blen
    uint32_t *s = &(wrk[3*blen]);   //1*blen
    uint32_t *r = &(wrk[4*blen]);   //1*blen
    uint32_t *wrki = &(wrk[5*blen]); //9*(blen+1)

    int cmp;
    int msib;
    int I;

    for(I=0;I<13*(blen+1);I++)    //init working memory to 0
        wrk[I] = 0;
    for(I=0;I<2*blen;I++)       //init lcm=0
        lcm[I] = 0;
    
    cmp = amsclil1_fb32_cmp(a,b,blen);  //bint is whichever is larger
    if(cmp>1)   //b>a
    {
        aint = a;
        bint = b;
    }
    else    //a>=b
    {
        aint = b;
        bint = a;
    }

    if(amsclil1_fb32_iszero(bint,blen))
    {
        //a = b = 0
        //set lcm to 0 (taken care of by default above)
        return;
    }

    msib = amsclil1_fb32_msi(bint,blen);
    //msib > msia by bint>aint


    //gcd = gcd(aint,bint)
    // actually requires 9*(blen+1)?
    amsclil1_fb32_eulergcd(aint,bint,gcd,wrki,msib);

    //s = bint / gcd
    // requires 5*(blen+1) working memory
    amsclil1_fb32_div(bint,gcd,q,r,wrki,msib);

    //lcm = aint*s
    // requires no working memory
    amsclil1_fb32_mult(aint,s,lcm,wrki,msib);

    return;
}

//Multiplicative Inverse Algorithm
//Refer to FIPS Pub 186-4, Algorithm C.1
//Inputs:
//  z [blen]    value to find the modular multiplicative inverse of
//  dom [blen]  modular domain 
//  wrk [17*(blen+1)]  internal working memory
//  blen - buffer length 
//Outputs:
//  zinv [blen]  multiplicative inverse such that mod(z*zinv,dom)==1
//Return:
//  1 - success, 0 - failure
int amsclil1_fb32_multinv(uint32_t *z, uint32_t *dom, uint32_t *zinv, uint32_t *wrk, int blen)
{
    int ret = 0;

    int cmp,res,res2;
    int I;

    uint32_t *xi = &(wrk[0]);        //blen
    uint32_t *xj = &(wrk[1*blen]);   //blen
    uint32_t *y = &(wrk[2*blen]);    //2*blen
    uint32_t *y1 = &(wrk[4*blen]);   //2*blen
    uint32_t *y2 = &(wrk[6*blen]);   //2*blen
    uint32_t *y3 = &(wrk[8*blen]);   //2*blen
    uint32_t *q = &(wrk[10*blen]);    //blen
    uint32_t *r = &(wrk[11*blen]);    //blen
    uint32_t *wrki = &(wrk[12*blen]); //div: [5*(blen+1)]
    //at least [17*(blen+1)]

    uint32_t *tmp = NULL;
    

    //An alternative way of handling this is take z --> mod(z,dom), then proceed

    for(I=0;I<blen;I++) zinv[I] = 0; //init zinv to 0

    //if z>=dom, error
    cmp = amsclil1_fb32_cmp(z,dom,blen);
    if(cmp!=1)
    {
        ret = 0;
        return ret;
    }

    for(I=0;I<17*(blen+1);I++) wrk[I] = 0; //clean work buffer to start with
    

    //xi = dom
    for(I=0;I<blen;I++) xi[I] = dom[I];
    //xj = z
    for(I=0;I<blen;I++) xj[I] = z[I];
    //y2 = 0
    for(I=0;I<blen;I++) y2[I] = 0;
    //y1 = 1
    for(I=0;I<blen;I++) y1[I] = 0; y1[0] = 1;

    res = amsclil1_fb32_iszero(xj,blen);
    while(res!=1)
    {
        //(q,r) = xi/xj
        res2 = amsclil1_fb32_div(xi,xj,q,r,wrki,blen);
        //y = y2-y1*q
        amsclil1_fb32_mult(y1,q,y3,wrki,blen);
        amsclil1_fb32_sub(y2,y3,y,blen);
        //xi = xj
        tmp = xi; xi = xj; xj = tmp;
        //xj = r
        tmp = xj; xj = r; r = tmp;
        //y2 = y1
        tmp = y2; y2 = y1; y1 = tmp;
        //y1 = y
        tmp = y1; y1=y; y=tmp;
    }

    //if xi != 1, failure
    res = amsclil1_fb32_msi(xi,blen);
    if(res!=1)
    {
        ret = 0;
        return ret;
    }
    if(xi[0]!=1)
    {
        ret = 0;
        return ret;
    }

    ret = 1;
    //zinv = y2, success
    for(I=0;I<blen;I++)
    {
        zinv[I] = y2[I];
    }

    return ret;
}
