#include <amsclil1/amsclil1.h>

#include <time.h>

//clock()
//CLOCKS_PER_SEC

//non_cryptographic test randomness
void amsclil1_fb32_testrandom(uint32_t* a,long blen)
{
    long I;
    
    for(I=0;I<blen;I++)
    {
        a[I] = (uint32_t) rand();
    }

    return;
}

void test_mult_time()
{
    long I;
    long N;
    uint32_t* a = NULL;
    uint32_t* b = NULL;
    uint32_t* c = NULL;
    uint32_t* d = NULL;
    uint32_t* wrk = NULL;

    double t0,t1;

    printf("amsclil1: Multiplication time test.\n");
    printf("bits \t time (msec)\n");
    printf("-----------------------------\n");
    for(I=1;I<=64;I++)
    {
        a = amsc_malloc(I*sizeof(uint32_t));
        b = amsc_malloc(I*sizeof(uint32_t));
        c = amsc_malloc(2*I*sizeof(uint32_t));
        d = amsc_malloc(I*sizeof(uint32_t));
        wrk = amsc_malloc(6*I*sizeof(uint32_t));
        amsclil1_fb32_testrandom(a,I);
        amsclil1_fb32_testrandom(b,I);
        amsclil1_fb32_testrandom(c,I);
        t0 = amsc_timemsec();
        amsclil1_fb32_mult(a,b,c,wrk,I);        
        t1 = amsc_timemsec();

        printf("%ld \t %1.4f\n",I*32,t1-t0);

        free(a); a = NULL;
        free(b); b = NULL;
        free(c); c = NULL;
        free(d); d = NULL;
    }

    return;
}

void test_div_time()
{
    printf("amsclil1: division time test.\n");
    long I;
    long N;
    uint32_t* a = NULL;
    uint32_t* b = NULL;
    uint32_t* c = NULL;
    uint32_t* d = NULL;
    uint32_t* wrk = NULL;

    double t0,t1;

    printf("bits \t time (msec)\n");
    printf("-----------------------------\n");
    for(I=1;I<=64;I++)
    {
        a = amsc_malloc(I*sizeof(uint32_t));
        b = amsc_malloc(I*sizeof(uint32_t));
        c = amsc_malloc(I*sizeof(uint32_t));
        d = amsc_malloc(I*sizeof(uint32_t));
        wrk = amsc_malloc(5*(I+1)*sizeof(uint32_t));
        amsclil1_fb32_testrandom(a,I);
        amsc_ui32_memset(b,0,I);
        amsclil1_fb32_testrandom(b,I/2);
        t0 = amsc_timemsec();
        amsclil1_fb32_div(a,b,c,d,wrk,I);        
        t1 = amsc_timemsec();

        printf("%ld \t %1.4f\n",I*32,t1-t0);

        free(a); a = NULL;
        free(b); b = NULL;
        free(c); c = NULL;
        free(d); d = NULL;
    }

    return;
}

void test_barrettmod()
{
    long N = 2;
    long K,Kx;
    uint32_t x[N],y[N],z[N],res[N],rem[N],mu[2*(N+1)];
    uint32_t wrk[33*(N+1)];
    uint32_t stor[14*(N+1)];

    printf("Barrett modulus tests.\n");

    printf("\n");
    amsc_ui32_memset(x,0,N); //x and y must have one leading zero digit
    amsc_ui32_memset(y,0,N);
    amsc_ui32_memset(mu,0,2*(N+1)); //because mu can have mis(y)+1 leading digit
    x[1] = 0x000000ab; x[0] = 0x12344321;
    y[1] = 0x00000001; y[0] = 0xf000000f;
    K = amsclil1_fb32_msi(y,N)+1;
    Kx = amsclil1_fb32_msi(x,N)+1;
    printf("K=%ld\n",K);
    printf("x="); amsclil1_fb32_liprint(x,N); printf("\n");
    printf("y="); amsclil1_fb32_liprint(y,N); printf("\n");
    amsclil1_fb32_barrettmu(y,mu,wrk,K,N);
    printf("mu="); amsclil1_fb32_liprint(mu,2*(N+1)); printf("\n");
    amsclil1_fb32_barrettmod(x,y,z,mu,wrk,K,Kx,N);
    printf("z=mod(x,y):"); amsclil1_fb32_liprint(z,N); printf("\n");
    //hangs with improper mu?

    // x=0x00000000000000ab12344321
    // y=0x0000000000000001f000000f
    // debug:mu=0x0000000000000000000000008421083e1194e97a4e97e82c
    // mu=0x8421083e1194e97a4e97e82c
    // z=mod(x,y):0x000000000000000092343df9
    //official answer: 92343df9 - okay, so that works

    printf("\n");
    amsc_ui32_memset(x,0,N); //x and y must have one leading zero digit
    amsc_ui32_memset(y,0,N);
    amsc_ui32_memset(mu,0,2*(N+1)); //because mu can have mis(y)+1 leading digit
    x[1] = 0x00000000; x[0] = 0x00000100;
    y[1] = 0x00000000; y[0] = 0x000000ff;
    K = amsclil1_fb32_msi(y,N)+1;
    Kx = amsclil1_fb32_msi(x,N)+1;
    printf("K=%ld\n",K);
    printf("x="); amsclil1_fb32_liprint(x,N); printf("\n");
    printf("y="); amsclil1_fb32_liprint(y,N); printf("\n");
    amsclil1_fb32_barrettmu(y,mu,wrk,K,N);
    printf("mu="); amsclil1_fb32_liprint(mu,2*(N+1)); printf("\n");
    amsclil1_fb32_barrettmod(x,y,z,mu,wrk,K,Kx,N);
    printf("z=mod(x,y):"); amsclil1_fb32_liprint(z,N); printf("\n");
    //answer should be 0x01

    printf("\n");
    amsc_ui32_memset(x,0,N); //x and y must have one leading zero digit
    amsc_ui32_memset(y,0,N);
    amsc_ui32_memset(mu,0,2*(N+1)); //because mu can have mis(y)+1 leading digit
    x[1] = 0x00000000; x[0] = 0x00000007;
    y[1] = 0x00000000; y[0] = 0x000000ff;
    K = amsclil1_fb32_msi(y,N)+1;
    Kx = amsclil1_fb32_msi(x,N)+1;
    printf("K=%ld\n",K);
    printf("x="); amsclil1_fb32_liprint(x,N); printf("\n");
    printf("y="); amsclil1_fb32_liprint(y,N); printf("\n");
    amsclil1_fb32_barrettmu(y,mu,wrk,K,N);
    printf("mu="); amsclil1_fb32_liprint(mu,2*(N+1)); printf("\n");
    amsclil1_fb32_barrettmod(x,y,z,mu,wrk,K,Kx,N);
    printf("z=mod(x,y):"); amsclil1_fb32_liprint(z,N); printf("\n");

    printf("\n");
    amsc_ui32_memset(x,0,N); //x and y must have one leading zero digit
    amsc_ui32_memset(y,0,N);
    amsc_ui32_memset(mu,0,2*(N+1)); //because mu can have mis(y)+1 leading digit
    x[1] = 0x00000000; x[0] = 0x00010000;
    y[1] = 0x00000000; y[0] = 0x0000ff00;
    K = amsclil1_fb32_msi(y,N)+1;
    Kx = amsclil1_fb32_msi(x,N)+1;
    printf("K=%ld\n",K);
    printf("x="); amsclil1_fb32_liprint(x,N); printf("\n");
    printf("y="); amsclil1_fb32_liprint(y,N); printf("\n");
    amsclil1_fb32_barrettmu(y,mu,wrk,K,N);
    printf("mu="); amsclil1_fb32_liprint(mu,2*(N+1)); printf("\n");
    amsclil1_fb32_barrettmod(x,y,z,mu,wrk,K,Kx,N);
    printf("z=mod(x,y):"); amsclil1_fb32_liprint(z,N); printf("\n");


    return;
}

void barret_modpower_test()
{
    long N = 2;
    long K,Kx;
    uint32_t x[N],y[N],z[N],res[N],rem[N],mu[2*(N+1)];
    uint32_t wrk[66*(N+1)];
    uint32_t stor[28*(N+1)];
    uint32_t a[N],b[N],c[N], d[N];


    printf("\nBarrett modpower tests.\n");

    printf("\n");
    amsc_ui32_memset(a,0,N);
    amsc_ui32_memset(b,0,N);
    amsc_ui32_memset(c,0,N);
    amsc_ui32_memset(d,0,N);
    a[0] = 0x00000010; a[1] = 0x00000000;
    b[0] = 0x00000002; b[1] = 0x00000000;
    c[0] = 0x000000ff; c[1] = 0x00000000;
    printf("mod(a^b, c) = d\n");
    printf("a="); amsclil1_fb32_liprint(a,N); printf("\n");
    printf("b="); amsclil1_fb32_liprint(b,N); printf("\n");
    printf("c="); amsclil1_fb32_liprint(c,N); printf("\n");
    amsclil1_fb32_modpower_cls(a,b,c,d,wrk,stor,N);
    printf("d="); amsclil1_fb32_liprint(d,N); printf("\n"); //0x01 - check
    amsclil1_fb32_modpower_bar(a,b,c,d,wrk,stor,N);
    printf("d2="); amsclil1_fb32_liprint(d,N); printf("\n"); 

    printf("\n");
    amsc_ui32_memset(a,0,N);
    amsc_ui32_memset(b,0,N);
    amsc_ui32_memset(c,0,N);
    amsc_ui32_memset(d,0,N);
    a[0] = 0x0000faf0; a[1] = 0x00000000;
    b[0] = 0x00000c0a; b[1] = 0x00000000;
    c[0] = 0x00ffaaff; c[1] = 0x00000000;
    printf("mod(a^b, c) = d\n");
    printf("a="); amsclil1_fb32_liprint(a,N); printf("\n");
    printf("b="); amsclil1_fb32_liprint(b,N); printf("\n");
    printf("c="); amsclil1_fb32_liprint(c,N); printf("\n");
    amsclil1_fb32_modpower_cls(a,b,c,d,wrk,stor,N);
    printf("d="); amsclil1_fb32_liprint(d,N); printf("\n"); //0xf5ee89 - check
    amsclil1_fb32_modpower_bar(a,b,c,d,wrk,stor,N);
    printf("d2="); amsclil1_fb32_liprint(d,N); printf("\n"); 

    printf("\n");
    amsc_ui32_memset(a,0,N);
    amsc_ui32_memset(b,0,N);
    amsc_ui32_memset(c,0,N);
    amsc_ui32_memset(d,0,N);
    a[0] = 0x0000afaf; a[1] = 0xafaf0000;
    b[0] = 0xff00ff00; b[1] = 0xff00ff00;
    c[0] = 0x10101010; c[1] = 0x10101010;
    printf("mod(a^b, c) = d\n");
    printf("a="); amsclil1_fb32_liprint(a,N); printf("\n");
    printf("b="); amsclil1_fb32_liprint(b,N); printf("\n");
    printf("c="); amsclil1_fb32_liprint(c,N); printf("\n");
    amsclil1_fb32_modpower_cls(a,b,c,d,wrk,stor,N);
    printf("d="); amsclil1_fb32_liprint(d,N); printf("\n"); //0x6e10c57c1e7bc71 - check
    amsclil1_fb32_modpower_bar(a,b,c,d,wrk,stor,N);
    printf("d2="); amsclil1_fb32_liprint(d,N); printf("\n"); 

    //speed test - moment of truth: can you do 4096 bit (64) RSA?
    long N2 = 64;
    uint32_t a2[N2],b2[N2],c2[N2],d2[N2];
    uint32_t wrk2[66*(N2+2)];
    uint32_t stor2[28*(N2+1)];

    printf("\n");
    amsc_ui32_memset(a2,0,N2);
    amsc_ui32_memset(b2,0,N2);
    amsc_ui32_memset(c2,0,N2);
    amsc_ui32_memset(d2,0,N2);
    a2[0] = 0x00000077; a2[N2-1] = 0x70000000;
    b2[0] = 0x00000066; b2[N2-1] = 0xff000000;
    c2[0] = 0x00000013; c2[N2-1] = 0x13000000;
    printf("mod(a^b, c) = d\n");
    printf("a2="); amsclil1_fb32_liprint(a2,N2); printf("\n");
    printf("b2="); amsclil1_fb32_liprint(b2,N2); printf("\n");
    printf("c2="); amsclil1_fb32_liprint(c2,N2); printf("\n");
    amsclil1_fb32_modpower_cls(a2,b2,c2,d2,wrk2,stor2,N2);
    printf("d2="); amsclil1_fb32_liprint(d2,N2); printf("\n"); 
    amsclil1_fb32_modpower_bar(a2,b2,c2,d2,wrk2,stor2,N2);
    printf("d2="); amsclil1_fb32_liprint(d2,N2); printf("\n"); 
 
}

void barrett_mod_fuzztest()
{
    long N = 32;
    long K,Kx;
    uint32_t x[N],y[N],z[N],res[N],rem[N],mu[2*(N+1)];
    uint32_t wrk[33*(N+1)];
    uint32_t stor[14*(N+1)];
    long I;
    int cmp;
    long Nfail;

    Nfail = 0;
    for(I=0;I<100;I++)
    {
        printf("Test %ld\n",I);
        amsc_ui32_memset(x,0,N);
        amsc_ui32_memset(y,0,N);
        amsc_ui32_memset(z,0,N);
        amsc_ui32_memset(res,0,N);
        amsc_ui32_memset(rem,0,N);
        amsc_ui32_memset(mu,0,2*(N+1));

        amsclil1_fb32_ncrandint(x,N);
        amsclil1_fb32_ncrandint(y,N/2);
        amsclil1_fb32_div(x,y,res,rem,wrk,N);

        K = amsclil1_fb32_msi(y,N)+1;
        Kx = amsclil1_fb32_msi(x,N)+1;
        amsclil1_fb32_barrettmu(y,mu,wrk,K,N);
        amsclil1_fb32_barrettmod(x,y,z,mu,wrk,K,Kx,N);
        
        cmp = amsclil1_fb32_cmp(z,rem,N);

        printf("  x="); amsclil1_fb32_liprint(x,N); printf("\n");
        printf("  y="); amsclil1_fb32_liprint(y,N); printf("\n");
        printf("rem="); amsclil1_fb32_liprint(rem,N); printf("\n");
        printf("  z="); amsclil1_fb32_liprint(z,N); printf("\n");

        if(cmp!=0)
        {
            printf("Mod accuracy failure:\n");
            printf("x="); amsclil1_fb32_liprint(x,N); printf("\n");
            printf("y="); amsclil1_fb32_liprint(y,N); printf("\n");
            printf("rem="); amsclil1_fb32_liprint(rem,N); printf("\n");
            printf("z="); amsclil1_fb32_liprint(z,N); printf("\n");
            Nfail++;
        }
    }
    printf("Nfailures = %ld out of %ld\n",Nfail,100L);
    
}


void modpower_timetests()
{
    printf("Modpower Time Tests.\n");

    long N;
    long I;
    uint32_t *x = NULL;
    uint32_t *y = NULL;
    uint32_t *md = NULL;
    uint32_t *pw = NULL;
    uint32_t *wrk = NULL;
    uint32_t *stor = NULL;

    double t0,t1;

    printf("bits \t time (msec)\n");
    printf("------------------");

    for(I=1;I<=64;I++)
    {
        N = I;
        x = amsc_malloc(I*sizeof(uint32_t));
        y = amsc_malloc(I*sizeof(uint32_t));
        md = amsc_malloc(I*sizeof(uint32_t));
        pw = amsc_malloc(I*sizeof(uint32_t));
        wrk = amsc_malloc(66*(I+1)*sizeof(uint32_t));
        stor = amsc_malloc(28*(I+1)*sizeof(uint32_t));

        amsclil1_fb32_testrandom(x,N);
        amsclil1_fb32_testrandom(md,N);
        amsclil1_fb32_testrandom(pw,N);

        t0 = amsc_timemsec();
        amsclil1_fb32_modpower(x,pw,md,y,wrk,stor,N);
        t1 = amsc_timemsec();
        printf("%ld \t %1.4f \n",N*32,t1-t0);

        free(x);    x=NULL;
        free(y);    y=NULL;
        free(md);   md=NULL;
        free(pw);   pw=NULL;
        free(wrk);  wrk=NULL;
        free(stor); stor=NULL;
    }


    return;
}

void test_amsclil1_fb32_eulergcd()
{
    printf("Test for amsclil1_fb32_eulergcd...\n");

    long blen = 6;
    uint32_t *a,*b,*c,*d,*e,*wrk;
    long I;
    int b1,b2;

    a = (uint32_t*) malloc(sizeof(uint32_t)*blen);
    b = (uint32_t*) malloc(sizeof(uint32_t)*blen);
    c = (uint32_t*) malloc(sizeof(uint32_t)*blen);
    d = (uint32_t*) malloc(sizeof(uint32_t)*blen);
    e = (uint32_t*) malloc(sizeof(uint32_t)*blen);
    wrk = (uint32_t*) malloc(10*sizeof(uint32_t)*(blen+1));
    
    for(I=0;I<1000;I++)
    {
        printf("euler gcd test %ld\n",I);
        amsclil1_fb32_ncrandint(a,blen);
        amsclil1_fb32_ncrandint(b,blen);

        //some sort of memory bug when blen>=6
        //still hitting bad conditions for division
        amsclil1_fb32_eulergcd(a,b,c,wrk,blen);
        amsclil1_fb32_div(a,c,d,e,wrk,blen);
        b1 = amsclil1_fb32_iszero(e,blen);
        amsclil1_fb32_div(b,c,d,e,wrk,blen);
        b2 = amsclil1_fb32_iszero(e,blen);
        if(b1==1 && b2 == 1)
        {
            printf("\t pass.\n");
        }
        else
        {
            printf("\t FAIL.\n");
            break;
        }
    }


    free(a);
    free(b);
    free(c);
    free(d);
    free(e);
    free(wrk);

    return;
}

void test_fb16_comparison_sub()
{
    printf("tests of fb16 subtraction and comparison ops.\n");
    long blen = 4;
    uint16_t *a,*b,*c,*wrk;
    long diff;
    int cmp;

    a = (uint16_t*) malloc(sizeof(uint16_t)*blen);
    b = (uint16_t*) malloc(sizeof(uint16_t)*blen);
    c = (uint16_t*) malloc(sizeof(uint16_t)*blen);

   //diff-1 = 0: a bug in subtraction 
// debug: 1 y2=0x00000000000000000000000000000000000000000000000000377d3a
// debug: 1 x2=0x00000000000000000000000000000000000000000000000000a62c4b
// debug: 1 y2=0x00000000000000000000000000000000000000000000000000377d3a
// debug: 1 x2=0x000000000000000000000000000000000000000000000000006eaf11
// debug: 1 y2=0x00000000000000000000000000000000000000000000000000377d3a
// debug: 1 x2=0x000000000000000000000000000000000000000000000000003731d7
// debug: 1 y2=0x00000000000000000000000000000000000000000000000000377d3a
// debug: 1 x2=0xffffffffffffffffffffffffffffffffffffffffffffffffffffb49d

    a[2] = 0;
    a[1] = 0x0037;
    a[0] = 0x7d3a;
    b[2] = 0;
    b[1] = 0x0037;
    b[0] = 0x31d7;

    printf("a="); amsclil1_fb16_liprint(a,blen); printf("\n");
    printf("b="); amsclil1_fb16_liprint(b,blen); printf("\n");
    
    diff = 0;
    cmp = amsclil1_fb16_intlscmp(a,b,diff,blen);
    printf("cmp(a,b) diff = %ld, cmp = %d\n",diff,cmp);

    diff = 0;
    cmp = amsclil1_fb16_intlscmp(b,a,diff,blen);
    printf("cmp(b,a) diff = %ld, cmp = %d\n",diff,cmp);

    diff = 1;
    cmp = amsclil1_fb16_intlscmp(a,b,diff,blen);
    printf("diff = %ld, cmp = %d\n",diff,cmp);

    diff = -1;
    cmp = amsclil1_fb16_intlscmp(a,b,diff,blen);
    printf("diff = %ld, cmp = %d\n",diff,cmp);

    diff = 2;
    cmp = amsclil1_fb16_intlscmp(a,b,diff,blen);
    printf("diff = %ld, cmp = %d\n",diff,cmp);

    printf("subtraction a - b*base^diff\n");
    diff = 0;
    printf("diff=%ld\n",diff);
    amsclil1_fb16_intlssub(a,b,diff,c,blen);
    printf("a="); amsclil1_fb16_liprint(a,blen); printf("\n");
    printf("b="); amsclil1_fb16_liprint(b,blen); printf("\n");
    printf("c="); amsclil1_fb16_liprint(c,blen); printf("\n");

    printf("subtraction a - b*base^diff\n");
    diff = 1;
    printf("diff=%ld\n",diff);
    amsclil1_fb16_intlssub(a,b,diff,c,blen);
    printf("a="); amsclil1_fb16_liprint(a,blen); printf("\n");
    printf("b="); amsclil1_fb16_liprint(b,blen); printf("\n");
    printf("c="); amsclil1_fb16_liprint(c,blen); printf("\n");

    // printf("subtraction a - b*base^diff\n");
    // diff = -1;
    // printf("diff=%ld\n",diff);
    // amsclil1_fb16_intlssub(a,b,diff,c,blen);
    // printf("a="); amsclil1_fb16_liprint(a,blen); printf("\n");
    // printf("b="); amsclil1_fb16_liprint(b,blen); printf("\n");
    // printf("c="); amsclil1_fb16_liprint(c,blen); printf("\n");

    printf("subtraction b - a*base^diff\n");
    diff = 0;
    printf("diff=%ld\n",diff);
    amsclil1_fb16_intlssub(b,a,diff,c,blen);
    printf("b="); amsclil1_fb16_liprint(b,blen); printf("\n");
    printf("a="); amsclil1_fb16_liprint(a,blen); printf("\n");
    printf("c="); amsclil1_fb16_liprint(c,blen); printf("\n");

    

    free(a);
    free(b);
    free(c);

    return;
}

void test_div_stress_test()
{
    uint32_t *a,*b,*q,*r,*m1,*m2, *wrk;
    long I;
    int res;
    long blen = 6;


    a = (uint32_t*) malloc(2*sizeof(uint32_t)*blen);
    b = (uint32_t*) malloc(2*sizeof(uint32_t)*blen);
    q = (uint32_t*) malloc(2*sizeof(uint32_t)*blen);
    r = (uint32_t*) malloc(2*sizeof(uint32_t)*blen);
    m1 = (uint32_t*) malloc(2*sizeof(uint32_t)*(blen+1));
    m2 = (uint32_t*) malloc(2*sizeof(uint32_t)*(blen+1));
    wrk = (uint32_t*) malloc(10*sizeof(uint32_t)*(blen+1));
    

    printf("Division stress test.\n");

    long ntest = 10;
    for(I=0;I<ntest;I++)
    {
        amsclil1_fb32_ncrandint(a,blen);
        amsclil1_fb32_ncrandint(b,blen-1);
        
        //a/b --> (q,r)
        amsclil1_fb32_div(a,b,q,r,wrk,blen);
        
        //q*b --> m1
        // Now I have a memory bug in the multiplication routine
        // probably m1 must be blen+1? (2*blen) - I wonder where else I've messed that up?
        amsclil1_fb32_mult(q,b,m1,wrk,blen);
        
        //m1+r --> m2
        amsclil1_fb32_add(m1,r,m2,2*blen);
        res = amsclil1_fb32_cmp(m2,a,2*blen);

        printf("test %ld: %d\n",I,res);
        printf("\ta= "); amsclil1_fb32_liprint(a,blen); printf("\n");
        printf("\tm2="); amsclil1_fb32_liprint(m2,blen); printf("\n");
        printf("\tb="); amsclil1_fb32_liprint(b,blen); printf("\n");
        printf("\tq="); amsclil1_fb32_liprint(q,blen); printf("\n");
        printf("\tr="); amsclil1_fb32_liprint(r,blen); printf("\n");
        
    }
    
    free(a);
    free(b);
    free(q);
    free(r);
    free(m1);
    free(m2);
    free(wrk);

    return;
}

int test_div_stress_test_sub(uint32_t *a, uint32_t *b, long blen)
{
    uint32_t *q,*r,*m1,*m2, *wrk;
    long I;
    int res,res2;

    q = (uint32_t*) malloc(2*sizeof(uint32_t)*blen);
    r = (uint32_t*) malloc(2*sizeof(uint32_t)*blen);
    m1 = (uint32_t*) malloc(2*sizeof(uint32_t)*(blen+1));
    m2 = (uint32_t*) malloc(2*sizeof(uint32_t)*(blen+1));
    wrk = (uint32_t*) malloc(10*sizeof(uint32_t)*(blen+1));

    amsclil1_fb32_setzero(q,2*blen);
    amsclil1_fb32_setzero(r,2*blen);
    amsclil1_fb32_setzero(m1,2*(blen+1));
    amsclil1_fb32_setzero(m2,2*(blen+1));
    amsclil1_fb32_setzero(wrk,10*(blen+1));

    //printf("\tm2="); amsclil1_fb32_liprint(m2,2*blen); printf("\n");

    //why is there extra shit in m2?

    //a/b --> (q,r)
    res2 = amsclil1_fb32_div(a,b,q,r,wrk,blen);
    
    //q*b --> m1
    // Now I have a memory bug in the multiplication routine
    // probably m1 must be blen+1? (2*blen) - I wonder where else I've messed that up?
    amsclil1_fb32_mult(q,b,m1,wrk,blen);
    
    //m1+r --> m2
    amsclil1_fb32_add(m1,r,m2,2*blen);
    res = amsclil1_fb32_cmp(m2,a,2*blen);

    if(res!=0 && res2!=0)
    {
        printf("Division test fail:\n");
        printf("cmp=%d  res=%d\n",res,res2);
        printf("\ta= "); amsclil1_fb32_liprint(a,2*blen); printf("\n");
        printf("\tm2="); amsclil1_fb32_liprint(m2,2*blen); printf("\n");
        printf("\tb="); amsclil1_fb32_liprint(b,blen); printf("\n");
        printf("\tq="); amsclil1_fb32_liprint(q,blen); printf("\n");
        printf("\tr="); amsclil1_fb32_liprint(r,blen); printf("\n");

        res = 0;
    }
    else
    {
        res = 1;
    }

    free(q);
    free(r);
    free(m1);
    free(m2);
    free(wrk);

    return res;
}

void test_div_stress_test2()
{
    long blenmax = 10;
    uint32_t *a,*b;

    long I,J,B,K,L;
    long Nt;
    int res;


    a = (uint32_t*) malloc(2*blenmax*sizeof(uint32_t));
    b = (uint32_t*) malloc(2*blenmax*sizeof(uint32_t));

    Nt = 1000; //number of tests at each division
    
    //test series 1
    printf("Test series 1: random divisions.\n");
    for(I=1;I<=blenmax;I++)
    {
        for(J=1;J<=blenmax;J++)
        {
            printf("\t divisions (%d,%d)\n",(int)I,(int)J);

            for(L=0;L<Nt;L++)
            {

                for(K=0;K<blenmax;K++)
                {
                    a[K] = 0;
                    b[K] = 0;
                }

                amsclil1_fb32_ncrandint(a,I);
                amsclil1_fb32_ncrandint(b,J);
                
                if(I>=J) B = I; else B = J;

                res = test_div_stress_test_sub(a,b,B);
                if(res<=0)
                {
                    printf("stopping on test failure...\n");
                    exit(0);
                }

            }
        }
    }

    printf("Test series 2: division by zero\n");
    for(I=1;I<=blenmax;I++)
    {
        for(J=1;J<=blenmax;J++)
        {
            printf("\t divisions (%d,%d)\n",(int)I,(int)J);

            for(L=0;L<Nt;L++)
            {

                for(K=0;K<blenmax;K++)
                {
                    a[K] = 0;
                    b[K] = 0;
                }

                amsclil1_fb32_ncrandint(a,I);
                //amsclil1_fb32_ncrandint(b,J);
                
                if(I>=J) B = I; else B = J;

                res = test_div_stress_test_sub(a,b,B);
                if(res<=0)
                {
                    printf("stopping on test failure...\n");
                    exit(0);
                }

            }
        }
    }


    printf("Test series 3: division of zero\n");
    for(I=1;I<=blenmax;I++)
    {
        for(J=1;J<=blenmax;J++)
        {
            printf("\t divisions (%d,%d)\n",(int)I,(int)J);

            for(L=0;L<Nt;L++)
            {

                for(K=0;K<blenmax;K++)
                {
                    a[K] = 0;
                    b[K] = 0;
                }

                //amsclil1_fb32_ncrandint(a,I);
                amsclil1_fb32_ncrandint(b,J);
                
                if(I>=J) B = I; else B = J;

                res = test_div_stress_test_sub(a,b,B);
                if(res<=0)
                {
                    printf("stopping on test failure...\n");
                    exit(0);
                }

            }
        }
    }


    printf("Test series 4: zero div zero\n");
    for(I=1;I<=blenmax;I++)
    {
        for(J=1;J<=blenmax;J++)
        {
            printf("\t divisions (%d,%d)\n",(int)I,(int)J);

            for(L=0;L<Nt;L++)
            {

                for(K=0;K<blenmax;K++)
                {
                    a[K] = 0;
                    b[K] = 0;
                }

                //amsclil1_fb32_ncrandint(a,I);
                //amsclil1_fb32_ncrandint(b,J);
                
                if(I>=J) B = I; else B = J;

                res = test_div_stress_test_sub(a,b,B);
                if(res<=0)
                {
                    printf("stopping on test failure...\n");
                    exit(0);
                }

            }
        }
    }


    free(a);
    free(b);

    return;
}

void test_ui64bitshifts()
{
    printf("Testing various aspects of uint64_t bit shifting...\n");

    uint64_t a,b,c,d;

    a = 0x01LL<<16;
    b = 0x01LL<<32;
    c = ((uint64_t)0x01)<<48;
    d = ((uint64_t)0x01)<<63;

    printf("%llx\n",(long long unsigned) a);
    printf("%llx\n",(long long unsigned) b);
    printf("%llx\n",(long long unsigned) c);
    printf("%llx\n",(long long unsigned) d);
    printf("%llx\n",(long long unsigned) (a+b+c+d));
    return;
}


static int amsclil1_test_fb32_lcm1_sub()
{
    int ret = 1;
    int N = 2048/32;
    uint32_t *a=NULL;
    uint32_t *b=NULL;
    uint32_t *c=NULL;
    uint32_t *q=NULL;
    uint32_t *r=NULL;
    uint32_t *wrk=NULL;
    
    int res1,res2;


    a = (uint32_t*) malloc(sizeof(uint32_t)*N);
    b = (uint32_t*) malloc(sizeof(uint32_t)*N);
    c = (uint32_t*) malloc(sizeof(uint32_t)*2*(N));
    q = (uint32_t*) malloc(sizeof(uint32_t)*2*N);
    r = (uint32_t*) malloc(sizeof(uint32_t)*2*N);
    wrk = (uint32_t*) malloc(sizeof(uint32_t)*14*(N+1));
    
    amsclil1_fb32_ncrandint(a,N);
    amsclil1_fb32_ncrandint(b,N);
    
    amsclil1_fb32_lcm(a,b,c,wrk,N);

    amsclil1_fb32_div(c,a,q,r,wrk,2*N);
    res1 = amsclil1_fb32_iszero(r,2*N);

    amsclil1_fb32_div(c,b,q,r,wrk,2*N);
    res2 = amsclil1_fb32_iszero(r,2*N);

    if(res1!=1 || res2!=1)
    {
        ret = 0;
    }

    free(a);
    free(b);
    free(c);
    free(q);
    free(r);
    free(wrk);

    
    return ret;
}

void amsclil1_test_fb32_lcm1()
{
    printf("Test LCM:\n\n");
    int I;
    int N = 1000;
    int ret;

    printf("Perofmring %d tests of LCM function\n",N);
    for(I=0;I<N;I++)
    {
       ret = amsclil1_test_fb32_lcm1_sub();
       if(ret!=1)
       {
           printf("LCM Tests failed at run %d of %d\n",I,N);
           break;
       }
    }
    if(ret==1)
    {
        printf("All tests passed.\n");
    }    

    return;
}