
#include <amsclil1/amsclil1.h>


// Unsigned Large Integer Comparison Operator: 
// 0 - a==b
// 1 - a>b
// 2 - a<b
int amsclil1_ulint_cmp(amsclil1_ulint *a, amsclil1_ulint *b)
{
    int ret = 0;
    int msia = amsclil1_ulint_msindex(a);
    int msib = amsclil1_ulint_msindex(b);
    int msimax;
    int I;
    uint32_t ad,bd;

    if(msia>msib) msimax = msia; else msimax = msib;

    for(I=msimax;I>=0;I--)
    {
        if(I<a->N) ad = a->data[I]; else ad = 0;
        if(I<b->N) bd = b->data[I]; else bd = 0;
        if(ad>bd)
        {
            ret = 1;
            break;
        }
        if(bd>ad)
        {
            ret = 2;
            break;
        }
    }

    return ret;
}

int amsclil1_ulint_iszero(amsclil1_ulint *a)
{
    int ret = 1;
    int I;
    for(I=a->N-1;I>=0;I--)
    {
        if(a->data[I]!=0)
        {
            ret = 0;
            break;
        }
    }
    return ret;
}

int amsclil1_ulint_add(amsclil1_ulint *a, amsclil1_ulint *b, amsclil1_ulint *c)
{
    int ret = 1;

    amsclil1_ulint *aint = NULL;
    amsclil1_ulint *bint = NULL;
    int msia,msib,msimax;

    if(c==NULL)
    {
        ret = 0;
        return ret;
    }

    msia = amsclil1_ulint_msindex(a);
    msib = amsclil1_ulint_msindex(b);
    if(msia>msib) msimax = msia; else msimax = msib;
    
    amsclil1_ulint_new(&aint);
    amsclil1_ulint_new(&bint);
    amsclil1_ulint_resize(aint,msimax+1);
    amsclil1_ulint_resize(bint,msimax+1);
    amsclil1_ulint_copy(aint,a);
    amsclil1_ulint_copy(bint,b);

    amsclil1_ulint_resize(c,msimax+1);
    amsclil1_ulint_clear(c);

    amsclil1_fb32_add(aint->data,bint->data,c->data,msimax+1);

    amsclil1_ulint_delete(&aint);
    amsclil1_ulint_delete(&bint);
    return ret;   
}

int amsclil1_ulint_sub(amsclil1_ulint *a, amsclil1_ulint *b, amsclil1_ulint *c)
{
    int ret = 1;

    amsclil1_ulint *aint = NULL;
    amsclil1_ulint *bint = NULL;
    int msia,msib,msimax;

    if(c==NULL)
    {
        ret = 0;
        return ret;
    }

    msia = amsclil1_ulint_msindex(a);
    msib = amsclil1_ulint_msindex(b);
    if(msia>msib) msimax = msia; else msimax = msib;
    
    amsclil1_ulint_new(&aint);
    amsclil1_ulint_new(&bint);
    amsclil1_ulint_resize(aint,msimax);
    amsclil1_ulint_resize(bint,msimax);
    amsclil1_ulint_copy(aint,a);
    amsclil1_ulint_copy(bint,b);

    amsclil1_ulint_resize(c,msimax);
    amsclil1_ulint_clear(c);

    amsclil1_fb32_sub(aint->data,bint->data,c->data,msimax);

    amsclil1_ulint_delete(&aint);
    amsclil1_ulint_delete(&bint);
    return ret;   
}

int amsclil1_ulint_mult(amsclil1_ulint *a, amsclil1_ulint *b, amsclil1_ulint *c)
{
    int ret = 1;

    amsclil1_ulint *aint = NULL;
    amsclil1_ulint *bint = NULL;
    int msia,msib,msimax;

    if(c==NULL)
    {
        ret = 0;
        return ret;
    }

    msia = amsclil1_ulint_msindex(a);
    msib = amsclil1_ulint_msindex(b);
    if(msia>msib) msimax = msia; else msimax = msib;
    
    amsclil1_ulint_new(&aint);
    amsclil1_ulint_new(&bint);
    amsclil1_ulint_resize(aint,msimax);
    amsclil1_ulint_resize(bint,msimax);
    amsclil1_ulint_copy(aint,a);
    amsclil1_ulint_copy(bint,b);

    amsclil1_ulint_resize(c,msimax);
    amsclil1_ulint_clear(c);

    amsclil1_fb32_sub(aint->data,bint->data,c->data,msimax);

    amsclil1_ulint_delete(&aint);
    amsclil1_ulint_delete(&bint);
    return ret;   
}